(* ::Package:: *)

(* Copyright (C) Khalil Ghorbal 2013-2015 *)
(* email. khalil.ghorbal@gmail.com *)
(* This package is released under GPL license. Please read the COPYRIGHT file *)

BeginPackage["InvariantGen`"]
ProjectiveVF::usage="HomogenizeVF[vf_,vars_] or HomogenizeVF[vf_,vars_,x0_] to specify a variable name to homogenize with"
GenProjectiveTemplate::usage="GenTemplate[d_,vars_,alpha_String]"
GenProjectiveEqSys::usage="GenEqSys[vf_,vars_,d_,N_,alpha_String,beta_String,x0_String]"
GenInvVar::usage="GenInvVar[vf_,vars_,d_,N_]"
Visitor::usage="Visitor[M_,obj_,m_Integer,symbols_List,sol_List]"
LieDerivative::usage="LieDerivative[p_,vf_, vars_List]"
MinRank::usage="MinRank[M_,symbols_List]"
SimplifyMatrix::usage="SimplifyMatrix[M_,beta_]"
Basis::usage="Basis[M_,eq_,symbols_]"
Basis1::usage="Basis1[M_,eq_,symbols_]"
WrapBenchmark::usage="WrapBenchmark[Benchmarks_List,maxdeg_,maxN_]"
RunBenchmarkInvVarGen::usage="RunBenchmarks[Benchmarks_List,maxdeg_,maxN_]"
orthogonalSpace::usage="orthogonal [v_]"
RemoveGammasPrimitive::usage="RemoveGammasPrimitive[oList_,vList_]"
RemoveGammas::usage="RemoveGammas[p_List,vars_List,symbols_List]"
ParallelInvVarietyGen::usage="ParallelInvVarietyGen[benchmark_,maxdeg_,maxN_,TC_]"
InvVarietyGen::usage="ParallelInvVarietyGen[benchmark_,maxdeg_,maxN_,TC_]"
GenEvolutionDomainConstraints::usage="GenEvolutionDomainConstraints[p_List,vars_,phi_]"
VectorFieldDegree::usage="VectorFieldDegree[f_List,vars_List]"
GenInvAlgebraicSets::usage="GenInvAlgebraicSets[benchmark_,phi_,maxdeg_,maxN_,TC_]"
PrettyPrint::usage="PrettyPrint[Sol_List,LieDerivatives_List,vars_,parameters_,L\[Alpha]_List,x0_String]"





(* The generation is performed in the projective space. There is no particular advantage in doing so. It only makes the program more symmetric and hence easier to write and to to understand  *)

Begin["`Private`"]
(* rewrite the list of vars into (x1,x2,...) *)

DEBUG=True;

LieDerivative[p_,vf_, vars_List]:=Module[{},
Grad[p,vars].vf
]

(* Returns -infty for the zero polynomial *)
PolynomialDegree[p_,vars_List]:=Module[{},
Max[Map[Plus @@ # &, CoefficientRules[p, vars][[All, 1]]]]
]

VectorFieldDegree[f_List,vars_List]:=Module[{},
Max[Map[PolynomialDegree[#, vars]&,f]]
]

ProjectiveVF[vf_,vars_,x0_String]:=Module[{tmp={},n=Length[vf],maxd=VectorFieldDegree[vf,vars],TotDegree=0},
Reap[Do[
tmp=CoefficientRules[vf[[i]], vars];
Do[TotDegree=Plus@@tmp[[j,1]];If[TotDegree<maxd, tmp=ReplacePart[tmp,{j->tmp[[j,1]]->tmp[[j,2]]*Symbol[x0]^(maxd-TotDegree)}]
],{j,1,Length[tmp]}];
Sow[FromCoefficientRules[tmp,vars]],{i,1,n}]
]][[2,1]] (* pick only the list built by Reap, no need for the rest *)

ProjectiveVF[vf_,vars_]:=Module[{},ProjectiveVF[vf,vars,"x0"]]

(*Generate (homogeneous) template; returns the template as well as the list of coefficients *)
GenProjectiveTemplate[d_,vars_,alpha_String,x0_String]:=Catch[Module[{n=Length[vars],m=0,L\[Alpha]={},tmp=0,L={},projvars=Join[vars,{Symbol[x0]}]},
If[d<0,Throw[{0,{}}]]; (* zero vector field *)
(*If[d===-Infinity,Throw[{0,{}}]]; 
If[d<0,Print["I can't generate a polynomial with negative degree. Aborting"];Abort[]];*)
m=Binomial[n+d,d];
L\[Alpha]=Flatten[Reap[Do[Sow[Symbol[StringJoin[alpha,ToString[i]]]],{i,m}]][[2]],1];
tmp=MonomialList[(Symbol[x0]+Sum[vars[[i]],{i,1,n}])^d,All,Lexicographic];
L=Flatten[CoefficientRules[tmp,projvars]];
Do[L=ReplacePart[L,{i->L[[i,1]]->L\[Alpha][[i]]}],{i,1,m}];
Throw[{FromCoefficientRules[L,projvars],L\[Alpha]}]
]]

GenProjectiveTemplate[d_,vars_,alpha_String]:=Module[{},GenProjectiveTemplate[d,vars,alpha,"x0"]]

GenProjectiveEqSys[vf_,vars_,d_,N_,alpha_String,beta_String,x0_String]:=Module[{p=0,hvf={},n=Length[vars],pdot=0,dvf=0,cofactors={},tmp=0,next=0,LieDerivatives={},L\[Alpha]={},EqSys={},M={}},
{p,L\[Alpha]}=GenProjectiveTemplate[d,vars,alpha];
Print[p];
dvf=VectorFieldDegree[vf,vars];
hvf=ProjectiveVF[vf,vars,x0];
tmp=p;
LieDerivatives=Reap[Do[next=LieDerivative[tmp,hvf,vars];Sow[next];tmp=next,{i,1,N}]][[2,1]];
PrependTo[LieDerivatives,p];
cofactors=Reap[Do[Sow[GenProjectiveTemplate[(N-i)*(-1+dvf),vars,StringJoin[beta,ToString[i]]]],{i,0,N-1}]][[2,1]];
EqSys=CoefficientRules[-LieDerivatives[[-1]]+Sum[cofactors[[i,1]]*LieDerivatives[[i]],{i,1,N}],Join[vars,{Symbol[x0]}]][[All,2]];
Print[EqSys];
If[EqSys==={},M={ConstantArray[0,Length[L\[Alpha]]]},M=CoefficientArrays[EqSys,L\[Alpha]][[2]]];
{M,LieDerivatives[[1;;-2]],L\[Alpha],Flatten[cofactors[[All,2]]]}
]

GenProjectiveEqSys[vf_,vars_,d_,N_]:=Module[{},GenProjectiveEqSys[vf,vars,d,N,"a","b","x0"]]

(**********************)
(** Basic Primitives **)
(**********************)

(* binary search, Compiled code *)
bsearchc=Compile[{{l,_Integer,1},{e,_Integer}},Module[{n0=1,n1=Length[l],len=Length[l],i=1,j=1},
While[n0<=n1,i=Floor[(n0+n1)/2];
If[l[[i]]==e,j=i;While[i>1&&l[[i-1]]==e,i--];While[j<len&&l[[j+1]]==e,j++];
Return[{i,j}]];
If[l[[i]]< e,n0=i+1,n1=i-1]];
If[l[[i]]<e,Return[{i,i}],Return[{i-1,i-1}]]],CompilationTarget->"C"]

(* Detect whether there exist either an upper or a lower triangular submatrix of the matrix matrules (encoded as a sparse array) *)
uploc=Compile[{{matrules,_Integer,2},{rows,_Integer,1},{m,_Integer}},Catch[Module[{rangeLO=Range[1,m],rangeUP=Range[1,m],j,k,tmp,
indices=matrules[[All,1]],pos,rowsUP={{0,0}},rowsLO={{0,0}},range(*,s=SessionTime[]*),l=0},
Do[
l=rows[[i]];
range=bsearchc[indices,l];
tmp=matrules[[range[[1]];;range[[2]],2]];
j=Max[tmp];
k=Min[tmp];
pos=Position[rangeUP,j,{1},1];
If[Length[pos]!=0,AppendTo[rowsUP,{l,j}];rangeUP=Delete[rangeUP,pos];];
pos=Position[rangeLO,k,{1},1];
If[Length[pos]!=0,AppendTo[rowsLO,{l,k}];rangeLO=Delete[rangeLO,pos];];
If[rangeUP==={},(*Print["up ",SessionTime[]-s];*)Throw[Drop[rowsUP,1]]];
If[rangeLO==={},(*Print["lo ",SessionTime[]-s];*)Throw[Drop[rowsLO,1]]];
,{i,Length[rows]}];
(*Print["none ",SessionTime[]-s];*)
Throw[{{0}}]]
],{{rowsUP,_Integer,2},{rowsLO,_Integer,2},{rangeUP,_Integer,1},{rangeLO,_Integer,1},{rows,_Integer,1},{range,_Integer,1},
{indices,_Integer,1},{tmp,_Integer,1},{pos,_Integer,2},{i,_Integer},{j,_Integer},{k,_Integer},{Position[rangeUP,j,{1},1],_Integer,2},
{Position[rangeLO,k,{1},1],_Integer,2},{Drop[rowsUP,1],_Integer,2},{Drop[rowsUP,1],_Integer,2}},CompilationTarget->"C"]

(* wrapping of uploc: returns 0 if no triangular submatrix exists, otherwise the det of the submatrix *)
uploeff[M_,matrules_,rows_,m_Integer]:=Catch[Module[{indices},
indices=uploc[matrules,rows,m];
If[indices==={{0}},Throw[0],Throw[Times@@Extract[M,indices]]];
]]

(* compute a basis for a given set of vectors *)
Basis1[M_,symbols_]:=Catch[Module[{rows={1},i=2,r=0,c=0,Lx={},names={},det,l={},perm={},permutedRange={}},
{r,c}=Dimensions[M];
Assert[r>=c];
(** Level[..,{-1}] is needed as variables alone will return Root[...] as a variable **)
names=Map[SymbolName,symbols];
Lx=Flatten[Reap[Do[Sow[Symbol[StringJoin["x",ToString[i]]]],{i,r}]][[2]],1];
(* Order the rows based on the number of their non-zero elements *)
Evaluate[symbols]=Evaluate[RandomInteger[{1,50},Length[symbols]]];
Do[
If[Reduce[Lx[[rows]].Normal[M[[rows]]]==Normal[M[[i]]],Lx[[rows]],Reals]===False,AppendTo[rows,i]];
If[Length[rows]==c,Break[]],{i,Range[1,r]}];
Map[Clear[Evaluate[#]]&,names];
(*Global`MM=M;Global`RR=rows;*)
If[Length[rows]==c,det=Det[M[[rows]]];Throw[det],Throw[0]]
]]

(* compute a basis for a given set of vectors *)
Basis[M_,symbols_]:=Catch[Module[{rows={},i=2,r=0,c=0,Lx={},names={},det,l={},perm={},permutedRange={}},
{r,c}=Dimensions[M];
Assert[r>=c];
(** Level[..,{-1}] is needed as variables alone will return Root[...] as a variable **)
names=Map[SymbolName,symbols];
Lx=Flatten[Reap[Do[Sow[Symbol[StringJoin["x",ToString[i]]]],{i,r}]][[2]],1];
(* Order the rows based on the number of their non-zero elements *)
perm=Ordering[Map[Length[ArrayRules[#]]-1&,M]];
permutedRange=Range[1,r][[perm]];
rows={permutedRange[[1]]};
Evaluate[symbols]=Evaluate[RandomInteger[{1,50},Length[symbols]]];
Do[
If[Reduce[Lx[[rows]].Normal[M[[rows]]]==Normal[M[[i]]],Lx[[rows]],Reals]===False,AppendTo[rows,i]];
If[Length[rows]==c,Break[]],{i,permutedRange}];
Map[Clear[Evaluate[#]]&,names];
(*Global`MM=M;Global`RR=rows;Global`perm=permutedRange;*)
If[Length[rows]==c,det=Det[M[[rows]]];Throw[det],Throw[0]]
]]

apply[M_,assign_]:=Catch[Module[{newM={},names={}},
If[assign==={},Throw[M],
names=Map[SymbolName,assign[[All,1]]];
Evaluate[assign[[All,1]]]=Evaluate[assign[[All,2]]]; (* much faster than applying a rule over the entire matrix *)
newM=M;
Map[Clear[Evaluate[#]]&,names];
Throw[newM]]
]]



(* simplify M knowing the exact value of some of its symbols. assign has the form "{symbol -> value, ...}" *)
(* Is there any better implementation for this ? *)
SimplifyMatrix[M_,assign_]:=Catch[Module[{ar={},rows={},columns={},newM={}},
newM=SparseArray[apply[M,assign]]; (* Removes the new zero elements but doesn't update the dimensions *)
ar=Drop[ArrayRules[newM],-1]; (* Drop {_,_}\[Rule]0 *)
If[ar==={},Throw[SparseArray[{{0}}]]];
rows=Sort[DeleteDuplicates[ar[[All,1,1]]]];
columns=Sort[DeleteDuplicates[ar[[All,1,2]]]];
Throw[newM[[rows]][[All,columns]]]
]]


ReducedToSetOfPoints[eq_List,symbols_]:=Catch[Module[{L={}},
(*Print["Prob:",eq];*)
(* Set a time out of 60s *)
TimeConstrained[L={ToRules[Reduce[eq,symbols,Reals]]},60,If[DEBUG,Print["Aborting the reduction."]];Throw[{False,eq}]];
(*Print["Sol:",L];*)
If[L==={}||MatchQ[Flatten[L[[All,All,2]]],{__?NumericQ}],Throw[{True,L}],Throw[{False,eq}]]
]]


(* Takes a pair {M,{assignements}} and returns a list of such pairs {{M1,{...}},{M2,{...}},...}  *)
VisitorFast[obj_,symbols_List]:=Catch[Module[{M=obj[[1]],assignedSymbols=obj[[2,All,1]],yetToAssign={},
det=0,l={},l1={},l2={},factors={},ar={},r=0,c=0,newM={}},

yetToAssign=Complement[symbols,assignedSymbols];
If[yetToAssign==={},Throw[{obj}]]; (* do nothing *)

If[obj[[3]]==={},
newM=M;,
l=FindInstance[obj[[3]],Intersection[Level[obj[[3]],{-1}],symbols],Reals,1];
(* if no instance is found or if all symbols are assigned, leave the matrix untouched for now, I have to think about it *)
If[l==={}||Length[l[[1]]]==Length[yetToAssign], Throw[{obj}] (* Do nothing for now, I need to think about it *),
  (*If[Reduce[obj[[3]],Intersection[Level[obj[[3]],{-1}],symbols],Reals]===False,Throw[{}],Print["FindInstance/Reduce mismatch, trying one more time ..."];newM=M];*)
   newM=SimplifyMatrix[M,l[[1]]]]
];

{r,c}=Dimensions[newM];
If[r<c,newM=Transpose[newM]];
ar=Drop[ArrayRules[newM],-1]; (* not really needed, will be removed *)
det=uploeff[newM,ar[[All,1]],Range[1,r],c];
If[PossibleZeroQ[det],det=Basis[newM,yetToAssign]];
(* what if det is again zero ? for now leave the matrix untouched, I have to think about it *)
If[PossibleZeroQ[det],Throw[{obj}]];

Assert[PolynomialQ[det,symbols]];
(*Print[det];*)
factors=Select[FactorList[det][[All,1]],Not[NumberQ[#]]&];
l=Map[ReducedToSetOfPoints[Join[{#==0},obj[[3]]],Intersection[Level[#,{-1}],yetToAssign]]&,factors];

l1=Select[l,#[[1]]==True&];
l2=Select[l,#[[1]]==False&];
If[l1=!={},l1=Flatten[l1[[All,2]],1]];
If[l2=!={},l2=l2[[All,2]]];

Throw[Join[Map[{SimplifyMatrix[M,#],Sort[Join[obj[[2]],#]],{}}&,l1],Map[{M,obj[[2]],Sort[#]}&,l2]]]
]]


(* Takes a pair {M,{assignements}} and returns a list of such pairs {{M1,{...}},{M2,{...}},...}  *)
Visitor0[obj_,symbols_List]:=Catch[Module[{M=obj[[1]],assignedSymbols=obj[[2,All,1]],yetToAssign={},
det=0,l={},l1={},l2={},factors={},ar={},r=0,c=0,newM={},Ly={}},

yetToAssign=Complement[symbols,assignedSymbols];
If[yetToAssign==={},Throw[{obj}]]; (* do nothing *)

If[obj[[3]]==={}
,
newM=M;
,
 If[Length[yetToAssign]==1
 ,
 (* univariate polynomial, span with all solution *)
 l={ToRules[Reduce[obj[[3]],yetToAssign,Reals]]}; (* returns a list with the added braces *)
 If[l==={},Throw[{}],Throw[Map[{SimplifyMatrix[M,#],Sort[Join[obj[[2]],#]],{}}&,l]]];
  ,
  l=FindInstance[obj[[3]],yetToAssign,Reals,1];
  If[l==={}
   ,
   If[Reduce[obj[[3]],yetToAssign,Reals]==False,Throw[{}],Print["Warning: FindInstance/Reduce mismatch."];Throw[{obj}]];
   ,
   If[Length[l[[1]]]==Length[yetToAssign]
    ,
   (* apply partially the solution with the hope that an additional constraint will appear *)
    newM=SimplifyMatrix[M,l[[1]][[2;;-1]]];
    ,
    newM=SimplifyMatrix[M,l[[1]]];
   ];
  ];
 ];
];
(*
  {r,c}=Dimensions[M];
  Ly=Flatten[Reap[Do[Sow[Symbol[StringJoin["y",ToString[i]]]],{i,c}]][[2]],1];
  rules=Reduce[Evaluate[ForAll[yetToAssign,And@@Map[#==0&,M.Ly]]],Join[yetToAssign,Ly],Reals];
  Print[MatrixForm[M]];
  Print[rules];
  If[And@@Map[#==0&,Ly]===rules,Throw[{}],Throw[{obj}]];
*)
{r,c}=Dimensions[newM];
If[r<c,newM=Transpose[newM]];
ar=Drop[ArrayRules[newM],-1]; (* not really needed, will be removed *)
det=uploeff[newM,ar[[All,1]],Range[1,r],c];
If[PossibleZeroQ[det],det=Basis[newM,yetToAssign]];
(* what if det is again zero ? for now leave the matrix untouched, I have to think about it *)
If[PossibleZeroQ[det],Throw[{obj}]];

Assert[PolynomialQ[det,symbols]];
(*Print[det];*)
factors=Select[FactorList[det][[All,1]],Not[NumberQ[#]]&];
l=Map[ReducedToSetOfPoints[Join[{#==0},obj[[3]]],Intersection[Level[#,{-1}],yetToAssign]]&,factors];

l1=Select[l,#[[1]]==True&];
l2=Select[l,#[[1]]==False&];
If[l1=!={},l1=Flatten[l1[[All,2]],1]];
If[l2=!={},l2=l2[[All,2]]];

Throw[Join[Map[{SimplifyMatrix[M,#],Sort[Join[obj[[2]],#]],{}}&,l1],Map[{M,obj[[2]],Sort[DeleteDuplicates[#]]}&,l2]]]
]]



(* Takes a pair {M,{assignements}} and returns a list of such pairs {{M1,{...}},{M2,{...}},...}  *)
Visitor[obj_,symbols_List]:=Catch[Module[{M=obj[[1]],assignedSymbols=obj[[2,All,1]],yetToAssign={},
det=0,l={},l1={},l2={},factors={},ar={},r=0,c=0,newM={},simplefactors={},nonsimplefactors={},sym={}},

yetToAssign=Complement[symbols,assignedSymbols];
If[yetToAssign==={},Throw[{obj}]]; (* do nothing *)
If[obj[[3]]==={False},Throw[{}]];

sym=Intersection[Level[obj[[3]],{-1}],yetToAssign];
If[Length[sym]==1,Throw[updateobjsimple[obj,obj[[3]],sym]]];

If[obj[[3]]==={}
,
newM=M;
,
  l=FindInstance[obj[[3]],yetToAssign,Reals,1];
  Assert[l=!={}];
   If[Length[l[[1]]]==Length[yetToAssign]
    ,
   (* apply partially the solution with the hope that an additional constraint will appear *)
    newM=SimplifyMatrix[M,l[[1]][[1;;-2]]];
    ,
    newM=SimplifyMatrix[M,l[[1]]];
   ];
];

{r,c}=Dimensions[newM];
If[r<c,newM=Transpose[newM]];
ar=Drop[ArrayRules[newM],-1]; (* not really needed, will be removed *)
det=uploeff[newM,ar[[All,1]],Range[1,r],c];
If[PossibleZeroQ[det],det=Basis[newM,yetToAssign]];
(* what if det is again zero ? for now leave the matrix untouched, I have to think about it *)
If[PossibleZeroQ[det],Throw[{obj}]];

Assert[PolynomialQ[det,symbols]];
(*Print[det];*)
factors=Select[FactorList[det][[All,1]],Not[NumberQ[#]]&];

simplefactors=Select[factors,Length[Variables[#]]==1&];
nonsimplefactors=Complement[factors,simplefactors];

Throw[
Flatten[
Join[
Map[updateobjsimple[obj,#==0,yetToAssign]&,simplefactors],
Map[updateobjnonsimple[obj,#==0,yetToAssign]&,nonsimplefactors]
],1]
]

]]



updateobjsimple[obj_,eq_,symbols_]:=Catch[Module[{M=obj[[1]],sol={},obj3={},ML={},list={},sym={}},
 sol={ToRules[Reduce[eq,symbols,Reals]]};
 If[sol==={},Throw[{}]];
 obj3=obj[[3]]/.sol/.{True}->{};
 ML=Map[SimplifyMatrix[M,#]&,sol];
 list=MapIndexed[{Extract[ML,#2],Sort[Join[obj[[2]],#1]],Extract[obj3,#2]}&,sol];
 Throw[list]
]]

updateobjnonsimple[obj_,eq_,symbols_]:=Catch[Module[{sol={},pb={},instances={}},
 pb=And@@Join[obj[[3]],{eq}];
 sol=Reduce[Exists[symbols,pb],symbols,Reals];
 If[sol==False,Throw[{}]];
 instances=FindInstance[pb,symbols,Reals,2];
 If[Length[instances]>1,Throw[{{obj[[1]],obj[[2]],Sort[DeleteDuplicates[Join[obj[[3]],{eq}]]]}}],If[DEBUG,Print["a sol exists but FindInstance cannot find it"]];Throw[{{obj[[1]],obj[[2]],Sort[DeleteDuplicates[Join[obj[[3]],{eq}]]]}}]];
 (* there exists mostlikely a unique solution *)
 sol=ToRules[Reduce[pb,symbols,Reals]];
 If[And@@Map[NumericQ,symbols/.sol]==True
  ,
  Throw[{{SimplifyMatrix[obj[[1]],sol],Sort[Join[obj[[2]],sol]],{}}}]
  ,
  Print["*** hm! ***"];Throw[{{obj[[1]],obj[[2]],Sort[DeleteDuplicates[Join[obj[[3]],{eq}]]]}}]
 ]
]]


(* Takes a symbolic matrix and maximizes the dimension of its kernel *)
MinRank[M_,symbols_List]:=Catch[Module[{current={{M,{},{}}},next={},K=Length[symbols],degeneratedCases={}},
Global`MM=M;
While[True,
 If[DEBUG,Print["# of objects: ",Length[current],". Path lengths ", Sort[DeleteDuplicates[Map[Length[#]&,current[[All,2]]]]],". Memory used (Gb): ",N[MemoryInUse[]/2^30]]];
 (*If[DEBUG,
 Print[TableForm[current]];
 Print["*** ***"];
 ];*) 
 next=DeleteDuplicates[Flatten[ParallelMap[Visitor[#,symbols]&,current],1]];
 If[next===current,Break[],current=next];
 (* an attempt to free some memory along the road *)
 ClearSystemCache[]; 
];
degeneratedCases=Select[current[[All,2]],Length[#]<K&];
If[degeneratedCases=!={},If[DEBUG,Global`DEGEN=degeneratedCases;Print["Degenerated cases occured"]]];
(*Throw[Select[current[[All,2]],Length[#]==K&]]*)
(* Always try the assignment \[Beta]i \[Rule] 0 by default *)

{r,c}=Dimensions[M];
hamiltonian=ToRules[symbols==0];
M0=apply[M,hamiltonian];
If[MatrixRank[M0]<c,Throw[Join[{hamiltonian},Select[current[[All,2]],Length[#]==K&]]],Throw[Select[current[[All,2]],Length[#]==K&]]];

]]


(* returns a basis for the space orthogonal to v *)
orthogonalSpace[v_]:=Module[{n=Length[v]},
Flatten[Reap[Do[Sow[PadRight[PadLeft[{-v[[i+1]],v[[i]]},i+1],n]],{i,n-1}]][[2]],1]
]


RemoveGammasPrimitive[oList_,vList_]:=
Catch[Module[{O={},v={},oo={}},
(*Print[MatrixForm[oList[[-1]]]];*)
If[vList=={},(*Print["I :: "];*)Throw[{Nest[Function[L,ReplacePart[L[[1;;-2]],-1->L[[-2]].L[[-1]]]],oList,-1+Length[oList]],{0}}]];
(*v=NullSpace[{vList[[1]]}.oList[[1]]];*)
oo=Nest[Function[L,ReplacePart[L[[2;;-1]],1->L[[1]].L[[2]]]],Join[{vList[[1]]},oList],Length[oList]][[1]];
v=orthogonalSpace[oo];
(*v=orthogonalSpace[vList[[1]].oList[[-1]]];*)
If[v=={},(*Print["II :: "];*)Throw[{Nest[Function[L,ReplacePart[L[[1;;-2]],-1->L[[-2]].L[[-1]]]],oList,-1+Length[oList]],Normal[oo]}]];
O=Transpose[v];
RemoveGammasPrimitive[Join[oList,{O}],vList[[2;;]]]
]]

RemoveGammas[p_List,vars_List,symbols_List]:=Catch[Module[{v0={},vars0={},L\[Gamma]={},L\[Nu]={},v={},A={},b={},initcondition={},M={}},
L\[Gamma]=Intersection[Level[p,{-1}],symbols];
{b,A}=CoefficientArrays[p,L\[Gamma]];
M=SimplifyMatrix[A,{}];
{r,c}=Dimensions[M];
Assert[r>0&&c>0];
(*v=NullSpace[{M[[1]]}];*)
If[DEBUG,Print["M :: ",MatrixForm[M]]];
If[c==1,conj=And@@Map[#==0&,Transpose[Normal[M]][[1]]];If[TrueQ[Reduce[Exists[vars,conj],vars,Reals]],Throw[conj],Throw[0]]];
v=orthogonalSpace[M[[1]]];
(*If[v==={},
 Print["HA! ",Normal[M]];
 Throw[And@@Normal[M[[1]]]==0],*)
 (*{v,initcondition}=RemoveGammasPrimitive[{Transpose[NullSpace[{M[[1]]}]]},M[[2;;]]];*)
 {v,initcondition}=RemoveGammasPrimitive[{Transpose[orthogonalSpace[M[[1]]]]},M[[2;;]]];
 If[DEBUG,Print["init ",initcondition]];
 If[DEBUG&&NumericQ[initcondition[[1]]]&&Not[PossibleZeroQ[initcondition[[1]]]],
   Print["Initial Condition is constant : ",initcondition," flushing"];Print[MatrixForm[NullSpace[M]]];Throw[{}]];
 (*If[PossibleZeroQ[initcondition[[1]]],initcondition={0}];*)
 If[DEBUG&&initcondition=!={0},Print["Conj ? ",MatrixForm[NullSpace[M]]]];
 v=Map[Transpose,v][[1]];
(*];*)
(* (TODO) What if initcondition is a constant. How to generate conjunctions ? *)
L\[Nu]=Flatten[Reap[Do[Sow[Symbol[StringJoin["\[Nu]",ToString[i]]]],{i,Length[L\[Gamma]]}]][[2]],1];
vars0=Flatten[Reap[Do[Sow[Symbol[StringJoin[ToString[vars[[i]]],"0"]]],{i,Length[vars]}]][[2]],1];
v0=apply[v,ToRules[vars==vars0]];
If[DEBUG,
Print[MatrixForm[M]];
Print["v0 :: ",v0]];
(*Print[v];
Print[v0];*)
(*/If[n==1,
 Throw[Join[initcondition/.ToRules[vars==vars0],{p[[1]]/.ToRules[L\[Gamma]==v0[[1]]]}]],
 (*Throw[Join[initcondition/.ToRules[vars==vars0],{p[[1]]/.ToRules[L\[Gamma]==L\[Nu][[1;;n]].v0]}]]*)
 Throw[Join[initcondition/.ToRules[vars==vars0],Map[p[[1]]/.ToRules[L\[Gamma]\[Equal]#]&,v0]]]
]*)
Throw[Implies[(initcondition[[1]]/.ToRules[vars==vars0])==0,And @@ Map[(p[[1]]/.ToRules[L\[Gamma]==#])==0&,v0]]]
]]

(* OLD IMPLEM

RemoveGammasPrimitive[M_]:=Catch[Module[{r=0,c=0,newM={},K=SimplifyMatrix[M,{}],O={}},
{r,c}=Dimensions[K];
Assert[r>=1];
If[c==1,Throw[{{},K}]];
O=orthogonalSpace[K[[1]]];
If[r==1,Throw[{O,K[[1]]}],
(* Update M *)
newM=Drop[K,1].Transpose[O];
RemoveGammasPrimitive[newM]
]
]]

RemoveGammas[p_List,vars_List,symbols_List]:=Catch[Module[{v0={},vars0={},L\[Gamma]={},L\[Nu]={},v={},A={},b={},n=0,initcondition={}},
L\[Gamma]=Intersection[Level[p,{-1}],symbols];
{b,A}=CoefficientArrays[p,L\[Gamma]];
{v,initcondition}=RemoveGammasPrimitive[A];
(* (TODO) I am not fully convinced that this is the only invariant we can get in this case. Think about this *)
If[v=={},Throw[Normal[initcondition]]];
L\[Nu]=Flatten[Reap[Do[Sow[Symbol[StringJoin["\[Nu]",ToString[i]]]],{i,Length[L\[Gamma]]}]][[2]],1];
vars0=Flatten[Reap[Do[Sow[Symbol[StringJoin[ToString[vars[[i]]],"0"]]],{i,Length[vars]}]][[2]],1];
n=Length[v];
v0=apply[v,ToRules[vars==vars0]];
If[n==1,
 Throw[p/.ToRules[L\[Gamma]==v0[[1]]]],
 Throw[p/.ToRules[L\[Gamma]==L\[Nu][[1;;n]].v0]]]
]]
*)


PrettyPrint[Sol_List,LieDerivatives_List,vars_,parameters_,L\[Alpha]_List,x0_String]:=Catch[Module[{L\[Gamma]={},nullSpaces={},GeneratorsL={},varsInit={},results={},nonTrivialResults={}},
nullSpaces=DeleteDuplicates[Map[NullSpace,Sol]];
L\[Gamma]=Flatten[Reap[Do[Sow[Symbol[StringJoin["\[Gamma]",ToString[i]]]],{i,Length[L\[Alpha]]}]][[2]],1];
varsInit=Map[Symbol[StringJoin[SymbolName[#],"0"]]&,vars];
GeneratorsL=Map[L\[Gamma][[1;;Length[#]]].#&,nullSpaces];
Set[Evaluate[Symbol[x0]],1];
(*Print[Map[LogicalExpand[Eliminate[{0\[Equal]p/.Join[ToRules[L\[Alpha]==#],ToRules[vars==varsInit]],0\[Equal]p/.ToRules[L\[Alpha]==#]},Intersection[Level[#,{-1}],L\[Gamma]]]]&,GeneratorsL]];*)
results=Map[LieDerivatives/.ToRules[L\[Alpha]==#]&,GeneratorsL];
Clear[Evaluate[x0]];
If[DEBUG,Print[results]];
nonTrivialResults=Select[results,Intersection[Level[#,{-1}],Complement[vars,parameters]]=!={}&];
(*
Print[L\[Gamma]];
Print[vars];
Print[nonTrivialResults];
*)
If[nonTrivialResults==={},Throw[{"No Non-Trivial Found"}],Throw[Map[RemoveGammas[#,vars,L\[Gamma]]&,nonTrivialResults]]]
]]

GenInvVar[vf_,vars_,d_,N_,alpha_String,beta_String,x0_String]:=Catch[Module[{ML={},L\[Alpha]={},L\[Beta]={},LieDerivatives={},Sol={},matrices={},parameters={}},
(*If[vf===ConstantArray[0,Length[vars]],Throw[{"Trivial VF"}]];*)
{ML,LieDerivatives,L\[Alpha],L\[Beta]}=GenProjectiveEqSys[vf,vars,d,N,alpha,beta,x0];
If[DEBUG,Print[Dimensions[ML]," with ",Length[L\[Beta]]," symbols"]];
Sol=MinRank[ML,L\[Beta]];
matrices=Map[apply[ML,#]&,Sol];
(* picks those variables equal to zero in the rhs *)
parameters=Complement[vars,Pick[vars,Replace[vf,x_/;Not[PossibleZeroQ[x]]->1,{1}],1]];
Throw[PrettyPrint[matrices,LieDerivatives,vars,parameters,L\[Alpha],x0]]
]]

GenInvVar[vf_,vars_,d_,N_]:=Module[{},CheckAbort[GenInvVar[vf,vars,d,N,"\[Alpha]","\[Beta]","x0"],If[DEBUG,Print["Abort Catched, continue computation"]];ClearSystemCache[];Clear["\[Beta]*"];Clear["\[Alpha]*"];Clear["\[Nu]*"];{}]]

WrapBenchmark[Benchmark_List,maxdeg_,maxN_]:=Module[{deg=1,order=1,normalized={},vars={},tmp={},list={},Lx={},wrapper={}},
vars=Benchmark[[2]];
Lx=Flatten[Reap[Do[Sow[Symbol[StringJoin["x",ToString[i]]]],{i,Length[vars]}]][[2]],1];
normalized=Benchmark/.ToRules[vars==Lx];
wrapper=Flatten[Reap[
For[deg=1,deg<=maxdeg,deg++,
For[order=1,order<=maxN,order++,
Sow[Join[normalized,{deg,order}]]
]
]][[2]],1];
{wrapper,vars,Lx}
]

WrapBenchmarkCAV2015[Benchmark_List,maxdeg_,maxN_]:=Module[{normalized={},vars={},tmp={},list={},Lx={},wrapper={}},
vars=Benchmark[[2]];
Lx=Flatten[Reap[Do[Sow[Symbol[StringJoin["x",ToString[i]]]],{i,Length[vars]}]][[2]],1];
normalized=Benchmark/.ToRules[vars==Lx];
(* CAV Benchmarks, {1,1},{2,1},{3,1},{4,1},{1,2} *)
(*wrapper={Join[normalized,{1,1}],Join[normalized,{1,2}],Join[normalized,{2,1}],Join[normalized,{3,1}],Join[normalized,{4,1}]};*)
wrapper={Join[normalized,{maxdeg,maxN}]};
{wrapper,vars,Lx}
]

RunBenchmarkInvVarGen[benchmark_,maxdeg_,maxN_,TC_]:=Module[{tab={},benchNormalized={},vars={},tmpvars={}},
Print[benchmark];
(* alpha renaming *)
{benchNormalized,vars,tmpvars}=WrapBenchmark[benchmark,maxdeg,maxN];
tab=Join[{{"","Trial.","d","N", "Invariants"}},
Table[Join[{"",i},{benchNormalized[[i]][[3]]},
{benchNormalized[[i]][[4]]},
TimeConstrained[GenInvVar[ReplacePart[benchNormalized[[i]],0->Sequence]],TC]/.ToRules[tmpvars==vars]/.{$Aborted->{StringForm[">``s",TC],Style["Timeout",White,Bold,Background->Gray]}}],{i,1,Length[benchNormalized]}]];
TableForm[tab]
]

(* Forces the Sow to be executed on the main Kernel
SetSharedFunction[ParallelSow];
ParallelSow[expr_] := Sow[expr]
*)

InvVarietyGen[benchmark_,maxdeg_,maxN_,TC_]:=Module[{tab={},vars={},tmpvars={},benchNormalized={},n=0,vars0={},tmpvars0={}},
If[DEBUG,Print[benchmark]];
(* alpha renaming *)
{benchNormalized,vars,tmpvars}=WrapBenchmarkCAV2015[benchmark,maxdeg,maxN];
n=Length[vars];
vars0=Flatten[Reap[Do[Sow[Subscript[vars[[i]],0]],{i,n}]][[2]],1];
tmpvars0=Flatten[Reap[Do[Sow[Symbol[StringJoin[ToString[tmpvars[[i]]],"0"]]],{i,n}]][[2]],1];

res=Select[Flatten[Map[TimeConstrained[GenInvVar[ReplacePart[#,0->Sequence]],TC,If[DEBUG,Print["TIMEOUT"]];Clear["\[Beta]*"];Clear["\[Alpha]*"];Clear["\[Gamma]*"];Clear["\[Nu]*"];0]&,benchNormalized]],Not[StringQ[#]]&]/.ToRules[tmpvars==vars]/.ToRules[tmpvars0==vars0];
res
(*Select[res,Intersection[Variables[#],vars0]==={}&]*)

(* FOR HSCC, simply replace vars0 by zero and pick the invariant polynomial function *)
(*
Select[Flatten[Map[TimeConstrained[GenInvVar[ReplacePart[#,0->Sequence]],TC,If[DEBUG,Print["TIMEOUT"]];Clear["\[Beta]*"];Clear["\[Alpha]*"];Clear["\[Gamma]*"];Clear["\[Nu]*"];0]&,benchNormalized]],Not[StringQ[#]]&]/.ToRules[tmpvars==vars]/.ToRules[tmpvars0==0]
*)
]

ParallelInvVarietyGen[benchmark_,maxdeg_,maxN_,TC_]:=Module[{tab={},vars={},tmpvars={},benchNormalized={},n=0,vars0={},tmpvars0={}},
If[DEBUG,Print[benchmark]];
(* alpha renaming *)
{benchNormalized,vars,tmpvars}=WrapBenchmark[benchmark,maxdeg,maxN];
n=Length[vars];
vars0=Flatten[Reap[Do[Sow[Subscript[vars[[i]],0]],{i,n}]][[2]],1];
tmpvars0=Flatten[Reap[Do[Sow[Symbol[StringJoin[ToString[tmpvars[[i]]],"0"]]],{i,n}]][[2]],1];
Select[Flatten[Map[TimeConstrained[GenInvVar[ReplacePart[#,0->Sequence]],TC,If[DEBUG,Print["TIMEOUT"]];Clear["\[Beta]*"];Clear["\[Alpha]*"];Clear["\[Gamma]*"];Clear["\[Nu]*"];0]&,benchNormalized]],Not[StringQ[#]]&]/.ToRules[tmpvars==vars]/.ToRules[tmpvars0==vars0]
]

GenEvolutionDomainConstraintsFromInv[p_,vars_,vars0_,phi_]:=Catch[Module[{phi0=True},
If[Intersection[Level[p,-1],vars0]==={},Throw[True],
phi0=phi/.ToRules[vars==vars0];
Throw[Resolve[Exists[vars0,p&&phi0],Reals]]
]
]]

GenEvolutionDomainConstraints[p_List,vars_,phi_]:=Module[{n=Length[vars],vars0={},familiesofInvVar={}},
vars0=Flatten[Reap[Do[Sow[Subscript[vars[[i]],0]],{i,n}]][[2]],1];
familiesofInvVar=Select[p,Intersection[Level[#,-1],vars0]=!={}&];
{And @@ ParallelMap[GenEvolutionDomainConstraintsFromInv[#,vars,vars0,phi]&,familiesofInvVar],Join[Select[familiesofInvVar/.ToRules[vars0==0],Not[TrueQ[#]]&],Complement[p,familiesofInvVar]]}
]

GenInvAlgebraicSets[benchmark_,phi_,maxdeg_,maxN_,TC_]:=Module[{vars=benchmark[[2]],p={},H=True,A={}},
p=InvVarietyGen[benchmark,maxdeg,maxN,TC];
If[DEBUG,Print["All Inv: ",p]];
{H,A}=GenEvolutionDomainConstraints[DeleteDuplicates[p],vars,phi];
{Simplify[H],DeleteDuplicates[A]}
]

End[ ]
EndPackage[ ]

(*

ClearSystemCache[];Clear["\[Beta]*"];Clear["\[Alpha]*"];GenInv[{x2,-x1+\[Epsilon]*(1-x1^2)*x2,0},{x1,x2,\[Epsilon]},2,1]//RuntimeTools`Profile

*)
